from kombu import Connection, Exchange, Queue

from twingly_pyamqp.amqp_config import (
    AMQPconfig,
)


class AmqpConnection:
    def __init__(
        self,
        config: AMQPconfig = None,
    ):
        self.config = config or AMQPconfig()
        self.connection = Connection(self.config.connection_url())

    def declare_queue(
        self,
        queue_name: str,
        exchange_name: str | None = None,
        routing_key: str | None = None,
        exchange_opts: dict | None = None,
        queue_opts: dict | None = None,
    ):
        if not exchange_name and routing_key:
            msg = "Default exchange cannot have a routing key"
            raise ValueError(msg)

        exchange = Exchange(exchange_name or "", **(exchange_opts or {}))
        queue = Queue(
            queue_name,
            exchange=exchange,
            routing_key=routing_key or queue_name,
            **(queue_opts or {}),
        )
        bound_queue = queue(self.connection)
        bound_queue.declare()

    def declare_exchange(
        self,
        exchange_name: str,
        exchange_opts: dict | None = None,
    ):
        exchange = Exchange(exchange_name, **(exchange_opts or {}))
        bound_exchange = exchange(self.connection)
        bound_exchange.declare()



from kombu import Exchange

from twingly_pyamqp.amqp_config import AMQPconfig
from twingly_pyamqp.amqp_connection import AmqpConnection


class Publisher(AmqpConnection):
    def __init__(
        self,
        exchange_name: str | None = None,
        routing_key: str | None = None,
        config: AMQPconfig = None,
    ):
        super().__init__(
            config=config,
        )
        self.exchange = Exchange(exchange_name)
        self.connection.routing_key = routing_key

    def publish(
        self,
        payload: object,
        routing_key: str | None = None,
        publish_args: dict | None = None,
    ):
        if not routing_key and not self.connection.routing_key:
            msg = "Routing key must be specified"
            raise ValueError(msg)

        self.connection.Producer(serializer="json").publish(
            payload,
            exchange=self.exchange,
            routing_key=routing_key or self.connection.routing_key,
            **(publish_args or {}),
        )



from .amqp_config import AMQPconfig
from .amqp_connection import AmqpConnection
from .publisher import Publisher
from .subscription import Subscription

__all__ = ["AMQPconfig", "AmqpConnection", "Publisher", "Subscription"]



import os

from dotenv import load_dotenv

load_dotenv()


class AMQPconfig:
    def __init__(
        self,
        rabbitmq_host: str | None = None,
        rabbitmq_port: int | None = None,
        amqp_user: str | None = None,
        amqp_password: str | None = None,
    ):
        self.rabbitmq_host = rabbitmq_host or os.getenv("RABBITMQ_N_HOST", "localhost")
        self.rabbitmq_port = rabbitmq_port or int(os.getenv("RABBITMQ_PORT", "5672"))
        self.amqp_user = amqp_user or os.getenv("AMQP_USERNAME", "guest")
        self.amqp_password = amqp_password or os.getenv("AMQP_PASSWORD", "guest")

    def connection_url(self):
        return f"amqp://{self.amqp_user}:{self.amqp_password}@{self.rabbitmq_host}:{self.rabbitmq_port}"



import contextlib
import threading
from collections.abc import Callable

from kombu import Queue

from twingly_pyamqp.amqp_config import AMQPconfig
from twingly_pyamqp.amqp_connection import AmqpConnection


class Subscription(AmqpConnection):
    def __init__(self, queue_names: str | list[str], config: AMQPconfig = None):
        self.consuming = threading.Event()
        self.thread = None
        self._thread_lock = threading.Lock()

        if isinstance(queue_names, str):
            queue_names = [queue_names]

        self.queues = [Queue(queue_name) for queue_name in queue_names]

        super().__init__(
            config=config,
        )

    def subscribe(
        self,
        callbacks: Callable[[str, object], None] | list[Callable[[str, object], None]],
        *,
        blocking: bool = True,
        timeout: int | None = None,
        consumer_args: dict | None = None,
    ) -> None:
        if not blocking and timeout is None:
            msg = "timeout must be specified for non-blocking mode"
            raise ValueError(msg)

        with self._thread_lock:
            if self.thread:
                msg = "Only one subscription can be started at a time, multiple queues can be consumed by a single subscription"
                raise ValueError(msg)

            self.consuming.set()
            if blocking:
                with contextlib.suppress(TimeoutError):
                    self._consume_messages(
                        callbacks, consumer_args, blocking=True, timeout=timeout
                    )

            else:
                self.thread = threading.Thread(
                    target=self._consume_messages,
                    args=(callbacks, consumer_args, False, timeout),
                )
                self.thread.daemon = True
                self.thread.start()

    def cancel(self) -> None:
        self.consuming.clear()
        with self._thread_lock:
            if self.thread:
                self.thread.join()
                self.thread = None

    def _consume_messages(self, callbacks, consumer_args, blocking, timeout=5) -> None:
        with self.connection.Consumer(
            self.queues,
            callbacks=callbacks if isinstance(callbacks, list) else [callbacks],
            **(consumer_args or {}),
        ) as _:
            while self.consuming.is_set():
                with (
                    contextlib.nullcontext()
                    if blocking
                    else contextlib.suppress(TimeoutError)
                ):
                    self.connection.drain_events(timeout=timeout)
